#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Db.h>

DbType DbTCharNew ();

/*---------------------------------------------------------------------------*/
char* DbCharReferenceGetAddress (DbReference* reference)
/*---------------------------------------------------------------------------*/
{
  DbObject object;
  char* result;

  object = DbReferenceGetObject (reference);

  if (!DbTypeIsChar (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbCharReferenceGetAddress>");
      return (NULL);
    }

  result = (char*) DbCharObjectGetAddress (object);
  return (result);
}

/*---------------------------------------------------------------------------*/
char* DbCharObjectLoad (char* name)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbLoadObject (name);
  if (DbGetError () != DbErrorOk) return (NULL);

  if (!DbTypeIsChar (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbCharObjectLoad>");
      return (NULL);
    }

  return ((char*) DbCharObjectGetAddress (object));
}

/*---------------------------------------------------------------------------*/
char* DbCharObjectLoadByTag (char* name, char* tag)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbLoadObjectByTag (name, tag);
  if (DbGetError () != DbErrorOk) return (NULL);

  if (!DbTypeIsChar (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbCharObjectLoad>");
      return (NULL);
    }

  return ((char*) DbCharObjectGetAddress (object));
}

/*---------------------------------------------------------------------------*/
char* DbCharObjectLoadByDate (char* name, char* date)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbLoadObjectByDate (name, date);
  if (DbGetError () != DbErrorOk) return (NULL);

  if (!DbTypeIsChar (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbCharObjectLoad>");
      return (NULL);
    }

  return ((char*) DbCharObjectGetAddress (object));
}

/*---------------------------------------------------------------------------*/
DbVector* DbCharVectorNew ()
/*---------------------------------------------------------------------------*/
{
  return (DbVectorNew (DbTCharNew ()));
}

/*---------------------------------------------------------------------------*/
void DbCharVectorInit (DbVector* vector)
/*---------------------------------------------------------------------------*/
{
  DbVectorInit (vector, DbTCharNew ());
}

/*---------------------------------------------------------------------------*/
char* DbCharVectorGetArray (DbVector* vector)
/*---------------------------------------------------------------------------*/
{
  char* result;
  DbType type;

  type = DbVectorGetElementType (vector);
  if (!DbTypeIsChar (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbCharVectorGetArray>");
      return (NULL);
    }

  result = (char*) DbVectorGetAddress (vector, 0);

  return (result);
}

/*---------------------------------------------------------------------------*/
char DbCharVectorGetElement (DbVector* vector, int index)
/*---------------------------------------------------------------------------*/
{
  char* result;
  DbType type;

  type = DbVectorGetElementType (vector);
  if (!DbTypeIsChar (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbCharVectorGetElement>");
      return (0);
    }

  if (index >= DbVectorGetSize (vector))
    {
      DbVectorExtend (vector, index + 1);
    }

  result = (char*) DbVectorGetAddress (vector, index);

  return (*result);
}

/*---------------------------------------------------------------------------*/
void DbCharVectorSetElement (DbVector* vector, 
                              int index, 
                              char value)
/*---------------------------------------------------------------------------*/
{
  char* result;
  DbType type;

  type = DbVectorGetElementType (vector);
  if (!DbTypeIsChar (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbCharVectorSetElement>");
      return;
    }

  if (index >= DbVectorGetSize (vector))
    {
      DbVectorExtend (vector, index + 1);
    }

  result = (char*) DbVectorGetAddress (vector, index);

  *result = value;
}

/*---------------------------------------------------------------------------*/
DbMatrix* DbCharMatrixNew (int rows, int columns)
/*---------------------------------------------------------------------------*/
{
  return (DbMatrixNew (DbTCharNew (), rows, columns));
}

/*---------------------------------------------------------------------------*/
void DbCharMatrixInit (DbMatrix* matrix, int rows, int columns)
/*---------------------------------------------------------------------------*/
{
  DbMatrixInit (matrix, DbTCharNew (), rows, columns);
}

/*---------------------------------------------------------------------------*/
char* DbCharMatrixGetArray (DbMatrix* matrix)
/*---------------------------------------------------------------------------*/
{
  char* result;
  DbType type;

  if (matrix == NULL) return (NULL);

  type = DbMatrixGetElementType (matrix);
  if (!DbTypeIsChar (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbCharMatrixGetArray>");
      return (NULL);
    }

  result = (char*) DbMatrixGetAddress (matrix, 0, 0);

  return (result);
}

/*---------------------------------------------------------------------------*/
char DbCharMatrixGetElement (DbMatrix* matrix, int row, int column)
/*---------------------------------------------------------------------------*/
{
  char* result;
  DbType type;

  if (matrix == NULL) return (0);

  type = DbMatrixGetElementType (matrix);
  if (!DbTypeIsChar (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbCharMatrixGetElement>");
      return (0);
    }

  result = (char*) DbMatrixGetAddress (matrix, row, column);

  return (*result);
}

/*---------------------------------------------------------------------------*/
void DbCharMatrixSetElement (DbMatrix* matrix, 
                              int row, 
                              int column, 
                              char value)
/*---------------------------------------------------------------------------*/
{
  char* result;
  DbType type;

  if (matrix == NULL) return;

  type = DbMatrixGetElementType (matrix);
  if (!DbTypeIsChar (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbCharMatrixSetElement>");
      return;
    }

  result = (char*) DbMatrixGetAddress (matrix, row, column);

  *result = value;
}

