#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Db.h>

DbType DbTDoubleNew ();

/*---------------------------------------------------------------------------*/
double* DbDoubleReferenceGetAddress (DbReference* reference)
/*---------------------------------------------------------------------------*/
{
  DbObject object;
  double* result;

  object = DbReferenceGetObject (reference);

  if (!DbTypeIsDouble (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbDoubleReferenceGetAddress>");
      return (NULL);
    }

  result = (double*) DbDoubleObjectGetAddress (object);
  return (result);
}

/*---------------------------------------------------------------------------*/
double* DbDoubleObjectLoad (char* name)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbLoadObject (name);
  if (DbGetError () != DbErrorOk) return (NULL);

  if (!DbTypeIsDouble (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbDoubleObjectLoad>");
      return (NULL);
    }

  return ((double*) DbDoubleObjectGetAddress (object));
}

/*---------------------------------------------------------------------------*/
double* DbDoubleObjectLoadByTag (char* name, char* tag)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbLoadObjectByTag (name, tag);
  if (DbGetError () != DbErrorOk) return (NULL);

  if (!DbTypeIsDouble (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbDoubleObjectLoad>");
      return (NULL);
    }

  return ((double*) DbDoubleObjectGetAddress (object));
}

/*---------------------------------------------------------------------------*/
double* DbDoubleObjectLoadByDate (char* name, char* date)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbLoadObjectByDate (name, date);
  if (DbGetError () != DbErrorOk) return (NULL);

  if (!DbTypeIsDouble (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbDoubleObjectLoad>");
      return (NULL);
    }

  return ((double*) DbDoubleObjectGetAddress (object));
}

/*---------------------------------------------------------------------------*/
DbVector* DbDoubleVectorNew ()
/*---------------------------------------------------------------------------*/
{
  return (DbVectorNew (DbTDoubleNew ()));
}

/*---------------------------------------------------------------------------*/
void DbDoubleVectorInit (DbVector* vector)
/*---------------------------------------------------------------------------*/
{
  DbVectorInit (vector, DbTDoubleNew ());
}

/*---------------------------------------------------------------------------*/
double* DbDoubleVectorGetArray (DbVector* vector)
/*---------------------------------------------------------------------------*/
{
  double* result;
  DbType type;

  type = DbVectorGetElementType (vector);
  if (!DbTypeIsDouble (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbDoubleVectorGetArray>");
      return (NULL);
    }

  result = (double*) DbVectorGetAddress (vector, 0);

  return (result);
}

/*---------------------------------------------------------------------------*/
double DbDoubleVectorGetElement (DbVector* vector, int index)
/*---------------------------------------------------------------------------*/
{
  double* result;
  DbType type;

  type = DbVectorGetElementType (vector);
  if (!DbTypeIsDouble (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbDoubleVectorGetElement>");
      return (0);
    }

  if (index >= DbVectorGetSize (vector))
    {
      DbVectorExtend (vector, index + 1);
    }

  result = (double*) DbVectorGetAddress (vector, index);

  return (*result);
}

/*---------------------------------------------------------------------------*/
void DbDoubleVectorSetElement (DbVector* vector, 
                              int index, 
                              double value)
/*---------------------------------------------------------------------------*/
{
  double* result;
  DbType type;

  type = DbVectorGetElementType (vector);
  if (!DbTypeIsDouble (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbDoubleVectorSetElement>");
      return;
    }

  if (index >= DbVectorGetSize (vector))
    {
      DbVectorExtend (vector, index + 1);
    }

  result = (double*) DbVectorGetAddress (vector, index);

  *result = value;
}

/*---------------------------------------------------------------------------*/
DbMatrix* DbDoubleMatrixNew (int rows, int columns)
/*---------------------------------------------------------------------------*/
{
  return (DbMatrixNew (DbTDoubleNew (), rows, columns));
}

/*---------------------------------------------------------------------------*/
void DbDoubleMatrixInit (DbMatrix* matrix, int rows, int columns)
/*---------------------------------------------------------------------------*/
{
  DbMatrixInit (matrix, DbTDoubleNew (), rows, columns);
}

/*---------------------------------------------------------------------------*/
double* DbDoubleMatrixGetArray (DbMatrix* matrix)
/*---------------------------------------------------------------------------*/
{
  double* result;
  DbType type;

  if (matrix == NULL) return (NULL);

  type = DbMatrixGetElementType (matrix);
  if (!DbTypeIsDouble (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbDoubleMatrixGetArray>");
      return (NULL);
    }

  result = (double*) DbMatrixGetAddress (matrix, 0, 0);

  return (result);
}

/*---------------------------------------------------------------------------*/
double DbDoubleMatrixGetElement (DbMatrix* matrix, int row, int column)
/*---------------------------------------------------------------------------*/
{
  double* result;
  DbType type;

  if (matrix == NULL) return (0);

  type = DbMatrixGetElementType (matrix);
  if (!DbTypeIsDouble (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbDoubleMatrixGetElement>");
      return (0);
    }

  result = (double*) DbMatrixGetAddress (matrix, row, column);

  return (*result);
}

/*---------------------------------------------------------------------------*/
void DbDoubleMatrixSetElement (DbMatrix* matrix, 
                              int row, 
                              int column, 
                              double value)
/*---------------------------------------------------------------------------*/
{
  double* result;
  DbType type;

  if (matrix == NULL) return;

  type = DbMatrixGetElementType (matrix);
  if (!DbTypeIsDouble (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbDoubleMatrixSetElement>");
      return;
    }

  result = (double*) DbMatrixGetAddress (matrix, row, column);

  *result = value;
}

