#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Db.h>

DbType DbTFloatNew ();

/*---------------------------------------------------------------------------*/
float* DbFloatReferenceGetAddress (DbReference* reference)
/*---------------------------------------------------------------------------*/
{
  DbObject object;
  float* result;

  object = DbReferenceGetObject (reference);

  if (!DbTypeIsFloat (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbFloatReferenceGetAddress>");
      return (NULL);
    }

  result = (float*) DbFloatObjectGetAddress (object);
  return (result);
}

/*---------------------------------------------------------------------------*/
float* DbFloatObjectLoad (char* name)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbLoadObject (name);
  if (DbGetError () != DbErrorOk) return (NULL);

  if (!DbTypeIsFloat (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbFloatObjectLoad>");
      return (NULL);
    }

  return ((float*) DbFloatObjectGetAddress (object));
}

/*---------------------------------------------------------------------------*/
float* DbFloatObjectLoadByTag (char* name, char* tag)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbLoadObjectByTag (name, tag);
  if (DbGetError () != DbErrorOk) return (NULL);

  if (!DbTypeIsFloat (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbFloatObjectLoad>");
      return (NULL);
    }

  return ((float*) DbFloatObjectGetAddress (object));
}

/*---------------------------------------------------------------------------*/
float* DbFloatObjectLoadByDate (char* name, char* date)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbLoadObjectByDate (name, date);
  if (DbGetError () != DbErrorOk) return (NULL);

  if (!DbTypeIsFloat (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbFloatObjectLoad>");
      return (NULL);
    }

  return ((float*) DbFloatObjectGetAddress (object));
}

/*---------------------------------------------------------------------------*/
DbVector* DbFloatVectorNew ()
/*---------------------------------------------------------------------------*/
{
  return (DbVectorNew (DbTFloatNew ()));
}

/*---------------------------------------------------------------------------*/
void DbFloatVectorInit (DbVector* vector)
/*---------------------------------------------------------------------------*/
{
  DbVectorInit (vector, DbTFloatNew ());
}

/*---------------------------------------------------------------------------*/
float* DbFloatVectorGetArray (DbVector* vector)
/*---------------------------------------------------------------------------*/
{
  float* result;
  DbType type;

  type = DbVectorGetElementType (vector);
  if (!DbTypeIsFloat (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbFloatVectorGetArray>");
      return (NULL);
    }

  result = (float*) DbVectorGetAddress (vector, 0);

  return (result);
}

/*---------------------------------------------------------------------------*/
float DbFloatVectorGetElement (DbVector* vector, int index)
/*---------------------------------------------------------------------------*/
{
  float* result;
  DbType type;

  type = DbVectorGetElementType (vector);
  if (!DbTypeIsFloat (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbFloatVectorGetElement>");
      return (0);
    }

  if (index >= DbVectorGetSize (vector))
    {
      DbVectorExtend (vector, index + 1);
    }

  result = (float*) DbVectorGetAddress (vector, index);

  return (*result);
}

/*---------------------------------------------------------------------------*/
void DbFloatVectorSetElement (DbVector* vector, 
                              int index, 
                              float value)
/*---------------------------------------------------------------------------*/
{
  float* result;
  DbType type;

  type = DbVectorGetElementType (vector);
  if (!DbTypeIsFloat (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbFloatVectorSetElement>");
      return;
    }

  if (index >= DbVectorGetSize (vector))
    {
      DbVectorExtend (vector, index + 1);
    }

  result = (float*) DbVectorGetAddress (vector, index);

  *result = value;
}

/*---------------------------------------------------------------------------*/
DbMatrix* DbFloatMatrixNew (int rows, int columns)
/*---------------------------------------------------------------------------*/
{
  return (DbMatrixNew (DbTFloatNew (), rows, columns));
}

/*---------------------------------------------------------------------------*/
void DbFloatMatrixInit (DbMatrix* matrix, int rows, int columns)
/*---------------------------------------------------------------------------*/
{
  DbMatrixInit (matrix, DbTFloatNew (), rows, columns);
}

/*---------------------------------------------------------------------------*/
float* DbFloatMatrixGetArray (DbMatrix* matrix)
/*---------------------------------------------------------------------------*/
{
  float* result;
  DbType type;

  if (matrix == NULL) return (NULL);

  type = DbMatrixGetElementType (matrix);
  if (!DbTypeIsFloat (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbFloatMatrixGetArray>");
      return (NULL);
    }

  result = (float*) DbMatrixGetAddress (matrix, 0, 0);

  return (result);
}

/*---------------------------------------------------------------------------*/
float DbFloatMatrixGetElement (DbMatrix* matrix, int row, int column)
/*---------------------------------------------------------------------------*/
{
  float* result;
  DbType type;

  if (matrix == NULL) return (0);

  type = DbMatrixGetElementType (matrix);
  if (!DbTypeIsFloat (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbFloatMatrixGetElement>");
      return (0);
    }

  result = (float*) DbMatrixGetAddress (matrix, row, column);

  return (*result);
}

/*---------------------------------------------------------------------------*/
void DbFloatMatrixSetElement (DbMatrix* matrix, 
                              int row, 
                              int column, 
                              float value)
/*---------------------------------------------------------------------------*/
{
  float* result;
  DbType type;

  if (matrix == NULL) return;

  type = DbMatrixGetElementType (matrix);
  if (!DbTypeIsFloat (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbFloatMatrixSetElement>");
      return;
    }

  result = (float*) DbMatrixGetAddress (matrix, row, column);

  *result = value;
}

