#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Db.h>

DbType DbTShortNew ();

/*---------------------------------------------------------------------------*/
short* DbShortReferenceGetAddress (DbReference* reference)
/*---------------------------------------------------------------------------*/
{
  DbObject object;
  short* result;

  object = DbReferenceGetObject (reference);

  if (!DbTypeIsShort (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbShortReferenceGetAddress>");
      return (NULL);
    }

  result = (short*) DbShortObjectGetAddress (object);
  return (result);
}

/*---------------------------------------------------------------------------*/
short* DbShortObjectLoad (char* name)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbLoadObject (name);
  if (DbGetError () != DbErrorOk) return (NULL);

  if (!DbTypeIsShort (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbShortObjectLoad>");
      return (NULL);
    }

  return ((short*) DbShortObjectGetAddress (object));
}

/*---------------------------------------------------------------------------*/
short* DbShortObjectLoadByTag (char* name, char* tag)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbLoadObjectByTag (name, tag);
  if (DbGetError () != DbErrorOk) return (NULL);

  if (!DbTypeIsShort (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbShortObjectLoad>");
      return (NULL);
    }

  return ((short*) DbShortObjectGetAddress (object));
}

/*---------------------------------------------------------------------------*/
short* DbShortObjectLoadByDate (char* name, char* date)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  object = DbLoadObjectByDate (name, date);
  if (DbGetError () != DbErrorOk) return (NULL);

  if (!DbTypeIsShort (DbObjectGetType (object)))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbShortObjectLoad>");
      return (NULL);
    }

  return ((short*) DbShortObjectGetAddress (object));
}

/*---------------------------------------------------------------------------*/
DbVector* DbShortVectorNew ()
/*---------------------------------------------------------------------------*/
{
  return (DbVectorNew (DbTShortNew ()));
}

/*---------------------------------------------------------------------------*/
void DbShortVectorInit (DbVector* vector)
/*---------------------------------------------------------------------------*/
{
  DbVectorInit (vector, DbTShortNew ());
}

/*---------------------------------------------------------------------------*/
short* DbShortVectorGetArray (DbVector* vector)
/*---------------------------------------------------------------------------*/
{
  short* result;
  DbType type;

  type = DbVectorGetElementType (vector);
  if (!DbTypeIsShort (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbShortVectorGetArray>");
      return (NULL);
    }

  result = (short*) DbVectorGetAddress (vector, 0);

  return (result);
}

/*---------------------------------------------------------------------------*/
short DbShortVectorGetElement (DbVector* vector, int index)
/*---------------------------------------------------------------------------*/
{
  short* result;
  DbType type;

  type = DbVectorGetElementType (vector);
  if (!DbTypeIsShort (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbShortVectorGetElement>");
      return (0);
    }

  if (index >= DbVectorGetSize (vector))
    {
      DbVectorExtend (vector, index + 1);
    }

  result = (short*) DbVectorGetAddress (vector, index);

  return (*result);
}

/*---------------------------------------------------------------------------*/
void DbShortVectorSetElement (DbVector* vector, 
                              int index, 
                              short value)
/*---------------------------------------------------------------------------*/
{
  short* result;
  DbType type;

  type = DbVectorGetElementType (vector);
  if (!DbTypeIsShort (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbShortVectorSetElement>");
      return;
    }

  if (index >= DbVectorGetSize (vector))
    {
      DbVectorExtend (vector, index + 1);
    }

  result = (short*) DbVectorGetAddress (vector, index);

  *result = value;
}

/*---------------------------------------------------------------------------*/
DbMatrix* DbShortMatrixNew (int rows, int columns)
/*---------------------------------------------------------------------------*/
{
  return (DbMatrixNew (DbTShortNew (), rows, columns));
}

/*---------------------------------------------------------------------------*/
void DbShortMatrixInit (DbMatrix* matrix, int rows, int columns)
/*---------------------------------------------------------------------------*/
{
  DbMatrixInit (matrix, DbTShortNew (), rows, columns);
}

/*---------------------------------------------------------------------------*/
short* DbShortMatrixGetArray (DbMatrix* matrix)
/*---------------------------------------------------------------------------*/
{
  short* result;
  DbType type;

  if (matrix == NULL) return (NULL);

  type = DbMatrixGetElementType (matrix);
  if (!DbTypeIsShort (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbShortMatrixGetArray>");
      return (NULL);
    }

  result = (short*) DbMatrixGetAddress (matrix, 0, 0);

  return (result);
}

/*---------------------------------------------------------------------------*/
short DbShortMatrixGetElement (DbMatrix* matrix, int row, int column)
/*---------------------------------------------------------------------------*/
{
  short* result;
  DbType type;

  if (matrix == NULL) return (0);

  type = DbMatrixGetElementType (matrix);
  if (!DbTypeIsShort (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbShortMatrixGetElement>");
      return (0);
    }

  result = (short*) DbMatrixGetAddress (matrix, row, column);

  return (*result);
}

/*---------------------------------------------------------------------------*/
void DbShortMatrixSetElement (DbMatrix* matrix, 
                              int row, 
                              int column, 
                              short value)
/*---------------------------------------------------------------------------*/
{
  short* result;
  DbType type;

  if (matrix == NULL) return;

  type = DbMatrixGetElementType (matrix);
  if (!DbTypeIsShort (type))
    {
      DbSetErrorInfo (DbErrorTypeMismatch, "DbShortMatrixSetElement>");
      return;
    }

  result = (short*) DbMatrixGetAddress (matrix, row, column);

  *result = value;
}

