/*---------------------------------------------------------------------------*/
/*        Copyright (c) 1996 LAL Orsay, UPS-IN2P3-CNRS (France).             */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions        */
/* are met:                                                                  */
/* 1. Redistributions of source code must retain the above copyright         */
/*    notice, this list of conditions and the following disclaimer.          */
/* 2. Redistributions in binary form must reproduce the above copyright      */
/*    notice, this list of conditions and the following disclaimer in the    */
/*    documentation and/or other materials provided with the distribution.   */
/* 3. All advertising materials mentioning features or use of this software  */
/*    must display the following acknowledgement:                            */
/*      This product includes software developed by the Computer Application */
/*      Development Group at LAL Orsay (Laboratoire de l'Accelerateur        */
/*      Linaire - UPS-IN2P3-CNRS).                                           */
/* 4. Neither the name of the Institute nor of the Laboratory may be used    */
/*    to endorse or promote products derived from this software without      */
/*    specific prior written permission.                                     */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE LAL AND CONTRIBUTORS ``AS IS'' AND       */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE     */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR        */
/* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE LAL OR CONTRIBUTORS BE      */
/* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR       */
/* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF      */
/* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS  */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN   */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)   */
/* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF    */
/* THE POSSIBILITY OF SUCH DAMAGE.                                           */
/*---------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <Db.h>
#include <DbObjectPath.h>
#include <CmMessage.h>

#include <DbCVS.h>
#include <DbLock.h>

#define tostring(x) #x

/*--------------------------------------------------------------------*/
typedef struct
{
  char* string;
  int allocated;
  int length;
} Buffer;

/*--------------------------------------------------------------------*/

/*--------------------------------------------------------------------*/

DbObject DbParserCreateObject (char* name);
DbObject DbParserLoadObjectCopy (char* name);
void DbParserLoadObjectCopies (char* text, CLList list);
DbType   DbParserCreateType (char* name);

/*--------------------------------------------------------------------*/

/*----- Handlers -----------------------------------------------------*/
#define DeclareMessageHandler(name) CmMessageStatus name (\
    CmMessage message, \
    char* sender, \
    char* server)


static DeclareMessageHandler (CreateTypeHandler);
static DeclareMessageHandler (LoadTypeHandler);
static DeclareMessageHandler (CreateObjectHandler);
static DeclareMessageHandler (CreateObjectsHandler);
static DeclareMessageHandler (LoadObjectHandler);
static DeclareMessageHandler (LoadObjectsHandler);
static DeclareMessageHandler (CommitObjectHandler);
static DeclareMessageHandler (CommitObjectsHandler);
static DeclareMessageHandler (CommitObjectAndHoldLockHandler);
static DeclareMessageHandler (LockObjectHandler);
static DeclareMessageHandler (UnlockObjectHandler);
static DeclareMessageHandler (AddTriggerHandler);
static DeclareMessageHandler (RemoveTriggerHandler);
static DeclareMessageHandler (DeleteObjectHandler);
static DeclareMessageHandler (DeleteTypeHandler);
static DeclareMessageHandler (UpdateTypeHandler);
static DeclareMessageHandler (UpdateHeadersHandler);
static DeclareMessageHandler (ObjectGetNamesHandler);
static DeclareMessageHandler (ObjectGetNamesInPathHandler);
static DeclareMessageHandler (SubPathGetNamesHandler);
static DeclareMessageHandler (TypeGetNamesHandler);
static DeclareMessageHandler (CommandHandler);
static DeclareMessageHandler (SetLocksHandler);
static DeclareMessageHandler (ResetLocksHandler);
static DeclareMessageHandler (GetVersionHandler);

static void DisconnectionHandler (CmConnect connect);
/*--------------------------------------------------------------------*/
static int SendVersion (char* dest);

/*--------------------------------------------------------------------*/
static void DbTStructureUpdateClients (DbTStructure structure);
static Buffer* BufferNew ();
static void BufferReset (Buffer* buffer);
static void BufferInitialize (Buffer* buffer);
static void BufferAdd (Buffer* buffer, char* text);

static DbObject DbLoadObjectFromFile (char* object_name, char* path_name);
static DbObject DbLoadObjectCopyFromFile (char* object_name, char* path_name);

int DbSendBreak (char* name, int transaction);

/*--------------------------------------------------------------------*/
static int      Finished = 0;
static Buffer*  Formatter = NULL;
static DbVersioningOption DefaultVersioningOption = DbUseVersioning;
static int      Transaction = 0;
static CmConnectConnectionHandler PreviousDisconnectionHandler = NULL;
/*--------------------------------------------------------------------*/

/*--------------------------------------------------------------------*/
int main (int argc, char* argv[])
/*--------------------------------------------------------------------*/
{
  char* server = DbServerName ();
  
  if (!CmMessageOpenServer (server))
    {
      fprintf (stderr, "Cm open server impossible\n");
      exit (1);
    }
  
  Formatter = BufferNew ();
  
  DbStartup ();
      
  DbInstallVersioningHandler ();

  CmMessageInstallHandler (CreateTypeHandler,           "DbCreateType");
  CmMessageInstallHandler (LoadTypeHandler,             "DbLoadType");
  CmMessageInstallHandler (CreateObjectHandler,         "DbCreateObject");
  CmMessageInstallHandler (CreateObjectsHandler,        "DbCreateObjects");
  CmMessageInstallHandler (LoadObjectHandler,           "DbLoadObject");
  CmMessageInstallHandler (LoadObjectsHandler,          "DbLoadObjects");
  CmMessageInstallHandler (CommitObjectHandler,         "DbCommitObject");
  CmMessageInstallHandler (CommitObjectsHandler,        "DbCiObjects");
  CmMessageInstallHandler (CommitObjectAndHoldLockHandler,
                           "DbCommitObjectAndHoldLock");
  CmMessageInstallHandler (LockObjectHandler,           "DbLockObject");
  CmMessageInstallHandler (UnlockObjectHandler,         "DbUnlockObject");
  CmMessageInstallHandler (AddTriggerHandler,           "DbAddTrigger");
  CmMessageInstallHandler (RemoveTriggerHandler,        "DbRemoveTrigger");
  CmMessageInstallHandler (DeleteObjectHandler,         "DbDeleteObject");
  CmMessageInstallHandler (DeleteTypeHandler,           "DbDeleteType");
  CmMessageInstallHandler (UpdateTypeHandler,           "DbUpdateType");
  CmMessageInstallHandler (UpdateHeadersHandler,        "DbUpdateHeaders");
  CmMessageInstallHandler (ObjectGetNamesHandler,       "DbObjectGetNames");
  CmMessageInstallHandler (ObjectGetNamesInPathHandler, "DbOGetNamesInPath");
  CmMessageInstallHandler (SubPathGetNamesHandler,      "DbSubPathGetNames");
  CmMessageInstallHandler (TypeGetNamesHandler,         "DbTypeGetNames");
  CmMessageInstallHandler (CommandHandler,              "DbCommand");
  CmMessageInstallHandler (SetLocksHandler,             "DbSetLocks");
  CmMessageInstallHandler (ResetLocksHandler,           "DbResetLocks");
  CmMessageInstallHandler (GetVersionHandler,           "DbGetVersion");
  
  CmServerSetComment ("ServerAlive");

  PreviousDisconnectionHandler = CmConnectInstallDisconnectionHandler (
      DisconnectionHandler);

  DbServerInstallVersioning ();

  DbClientRestoreLocksAndTriggers ();

  Transaction = CmOpenTransaction ("DbServer", NULL);

  while (!Finished)
    {
      CmRestartTransaction (Transaction);
      while (!CmIsTransactionTerminated (Transaction))
        {
          CmMessageWait ();
        }
    }

  CmCloseTransaction (Transaction);
  
  DbVersioningCleanup ();
  DbCleanup ();

  return (Finished);
}

/*---------------------------------------------------------------------------*/
char* BuildFullName (char* object_name, char* path_name)
/*---------------------------------------------------------------------------*/
{
  static char full_name[512];
  
  if ((path_name != NULL) &&
      (strlen (path_name) > 0))
    {
      strcpy (full_name, path_name);
      strcat (full_name, "/");
    }
  else
    {
      full_name[0] = 0;
    }
  
  strcat (full_name, object_name);
  
  return (full_name);
}

/*---------------------------------------------------------------------------*/
char* BuildFileName (char* object_name, char* path_name)
/*---------------------------------------------------------------------------*/
{
  static char file_name[512];

  strcpy (file_name, BuildFullName (object_name, path_name));
  strcat (file_name, ".db");
  
  return (file_name);
}

/*---------------------------------------------------------------------------*/
char* BuildTmpFileName (char* object_name, char* path_name, char* suffix)
/*---------------------------------------------------------------------------*/
{
  static char file_name[512];
  
  strcpy (file_name, BuildFileName (object_name, path_name));
  strcat (file_name, ".");
  strcat (file_name, suffix);

  return (file_name);
}

/*---------------------------------------------------------------------------*/
static int BuildHeaderFile (DbType type)
/*---------------------------------------------------------------------------*/
{
  DbTStructure structure;
  CIterator it;
  char* name;
  char command[256];
  char file_name[256];
  FILE* file;
  int test = 0;
  
  if (!DbTypeIsStructure (type)) return (0);
  
  structure = (DbTStructure) type;
  name = DbTypeName (type);
  
  sprintf (file_name, "%s.h", name);
  
  sprintf (command, "sed 's/<T>/%s/g' ${DBROOT}/src/DbTemplate1.h >%s",
           name, file_name);
  
  system (command);
  

  /* mettre ici la partie qui repere la presence d une structure "maison"
     inclure dans le fichier la string: #include "STRUCT.h" 
  */

  file = fopen (file_name, "a+");
  if (file != NULL)
    {
      fprintf (file, "typedef struct\n");
      fprintf (file, "{\n");
      
      it = DbTStructureGiveFieldIterator (structure);
      
      for (; !CIteratorFinished (it); CIteratorForth (it))
        {
          DbField field;
          DbType type;
          char* type_name;
          
          field = (DbField) CIteratorItem (it);
          
          type = DbFieldType (field);
          
          type_name = DbTypeCName (type);
          
          fprintf (file, "  %s %s;\n", 
                   type_name,
                   DbFieldGetName (field));
	  test++;
        }

      if (test == 0)
	DbSetErrorInfo (DbErrorBadDimension,
			"Check your object field number (maximum = 50)");
      
      CIteratorDelete (it);
      
      fprintf (file, "} %s;\n", name);
      
      fclose (file);
    }
  
  sprintf (command, "sed 's/<T>/%s/g' ${DBROOT}/src/DbTemplate2.h >>%s",
           name, file_name);
  
  system (command);
  
  return (1);
}

/*---------------------------------------------------------------------------*/
static int BuildCppHeaderFile (DbType type)
/*---------------------------------------------------------------------------*/
{
  DbTStructure structure;
  CIterator it;
  char* name;
  char command[256];
  char file_name[256];
  FILE* file;
  
  if (!DbTypeIsStructure (type)) return (0);
  
  structure = (DbTStructure) type;
  name = DbTypeName (type);
  
  sprintf (file_name, "%s.hxx", name);
  
  sprintf (command, "sed 's/<T>/%s/g' ${DBROOT}/src/DbTemplate1.hxx >%s",
           name, file_name);
  
  system (command);
  
  file = fopen (file_name, "a+");
  if (file != NULL)
    {
      it = DbTStructureGiveFieldIterator (structure);
      
      for (; !CIteratorFinished (it); CIteratorForth (it))
        {
          DbField field;
          DbType type;
          char* type_name;
          
          field = (DbField) CIteratorItem (it);
          
          type = DbFieldType (field);
          
          type_name = DbTypeCppName (type);

          fprintf (file, "  %s %s;\n", 
                   type_name,
                   DbFieldGetName (field));
        }
      
      CIteratorDelete (it);
      
      fclose (file);
    }
  
  sprintf (command, "sed 's/<T>/%s/g' ${DBROOT}/src/DbTemplate2.hxx >>%s",
           name, file_name);
  
  system (command);
  
  return (1);
}

/*---------------------------------------------------------------------------*/
static char* FormatType (DbType type)
/*---------------------------------------------------------------------------*/
{
  DbTStructure structure;
  CIterator it;
  char* name;
  
  if (!DbTypeIsStructure (type)) return (NULL);
  
  structure = (DbTStructure) type;
  name = DbTypeName (type);
  
  BufferReset (Formatter);
  
  BufferAdd (Formatter, "structure ");
  BufferAdd (Formatter, name);
  BufferAdd (Formatter, " { ");
  
  it = DbTStructureGiveFieldIterator (structure);
  
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbField field;
      DbType type;
      char* type_name;
      
      field = (DbField) CIteratorItem (it);
      
      type = DbFieldType (field);
      
      type_name = DbTypeName (type);
      
      BufferAdd (Formatter, type_name);
      BufferAdd (Formatter, " ");
      BufferAdd (Formatter, DbFieldGetName (field));
      BufferAdd (Formatter, " ");
    }
  
  CIteratorDelete (it);
  
  BufferAdd (Formatter, "}");
  
  return (Formatter->string);
}

/*--------------------------------------------------------------------*/
static void DbTStructureUpdateClients (DbTStructure structure)
/*--------------------------------------------------------------------*/
{
  CIterator it;
  FILE* file;
  char file_name[256];
  
  it = DbTStructureGiveFieldIterator (structure);
  
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbField field;
      DbType type;
      
      field = (DbField) CIteratorItem (it);
      
      type = DbFieldType (field);
      type = DbTypeGetUsedStructure (type);
      if (type != NULL)
        {
          DbTStructureAddClient (structure, type);
        }
    }
  
  CIteratorDelete (it);
}

/*---------------------------------------------------------------------*/
static DbType LoadType (char* name)
/*---------------------------------------------------------------------*/
{
  DbType type = NULL;
  char type_name[256];
  char file_name[256];
  char* text;
  
  if ((name == NULL) ||
      (strlen (name) == 0))
    {
      DbSetErrorInfo (DbErrorBadArgument, "Server(LoadType)");
      return (NULL);
    }
  
  strcpy (type_name, name);
  
  for (;;)
    {
      DbClearErrorInfo ();
      
      sprintf (file_name, "%s.ddl", type_name);
      text = DbReadFile (file_name);
      if (text != NULL)
        {
          GCReferenceAssign (type, DbParserCreateType (text));
          free (text);
          
          if (DbGetError () == DbErrorUnloadedStructure)
          {
            DbTStructureUninstall (type);
            GCReferenceClear (type);

            if (LoadType (DbGetErrorInfo ()) == NULL) break;
          }
          else
          {
            if (type != NULL)
            {
              DbTStructure structure = (DbTStructure) type;
              
              DbTStructureUpdateClients (structure);
            }
            break;
          }
        }
      else
        {
          DbSetErrorInfo (DbErrorUnknownStructure, "%s", file_name);
          break;
        }
    }

  return (type);
}

/*--------------------------------------------------------------------*/
DbObject DbLoadObjectFromText (char* text)
/*--------------------------------------------------------------------*/
{
  DbObject object = NULL;
  
  if (text == NULL) return (NULL);
      
  for (;;)
    {
      DbClearErrorInfo ();
      
      object = DbParserCreateObject (text);
      
      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (LoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
  
  return (object);
}

/*--------------------------------------------------------------------*/
DbObject DbLoadObjectCopyFromText (char* text)
/*--------------------------------------------------------------------*/
{
  DbObject object = NULL;
  
  if (text == NULL) return (NULL);
      
  for (;;)
    {
      DbClearErrorInfo ();
      
      object = DbParserLoadObjectCopy (text);
      
      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (LoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
  
  return (object);
}

/*--------------------------------------------------------------------*/
int DbLoadObjectCopiesFromText (char* text, CLList list)
/*--------------------------------------------------------------------*/
{
  DbObject object = NULL;
  
  if (text == NULL) return (0);
  if (list == NULL) return (0);
      
  for (;;)
    {
      DbClearErrorInfo ();

      CLListErase (list);
      
      DbParserLoadObjectCopies (text, list);
      
      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (LoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
  
  return (CLListCount (list));
}

/*--------------------------------------------------------------------*/
static DbObject DbLoadObjectFromFile (char* object_name, char* path_name)
/*--------------------------------------------------------------------*/
{
  DbObject  object = NULL;
  char*     file_name;
  char*     text;
  
  file_name = BuildFileName (object_name, path_name);
  

  text = DbReadFile (file_name);
  if (text != NULL)
    {
      object = DbLoadObjectFromText (text);
      free (text);
    }
  else
    {
      DbSetErrorInfo (DbErrorUnknownObject, "LoadObjectFromFile>");
    }
  
  return (object);
}

/*--------------------------------------------------------------------*/
static DbObject DbLoadObjectCopyFromFile (char* object_name, char* path_name)
/*--------------------------------------------------------------------*/
{
  DbObject  object = NULL;
  char*     file_name;
  char*     text;
  
  file_name = BuildFileName (object_name, path_name);
  
  text = DbReadFile (file_name);
  if (text != NULL)
    {
      object = DbLoadObjectCopyFromText (text);
      free (text);
    }
  else
    {
      DbSetErrorInfo (DbErrorUnknownObject, "LoadObjectCopyFromFile>");
    }
  
  return (object);
}

/*----- Handlers -----------------------------------------------------*/

/*--------------------------------------------------------------------*/
static CmMessageStatus CreateTypeHandler (CmMessage message, 
                                          char* sender,
                                          char* server)
/*--------------------------------------------------------------------*/
{
  char* text;
  int transaction = 0;

  text = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("CreateTypeHandler> from %s text = %s\n", sender, text);
    }
  
  if (text != NULL)
    {
      DbType type;
      
      DbClearErrorInfo ();
      
      type = DbParserCreateType (text);
      
      if (type == NULL)
        {
          DbSetErrorInfo (DbErrorBadSyntax, "");
        }
      else if (!DbTypeIsStructure (type))
        {
          DbSetErrorInfo (DbErrorNotStructure, 
                          "Type %s is not a structure", DbTypeName (type));
        }
      else
        {
          char file_name[256];

          sprintf (file_name, "%s.ddl", DbTypeName (type));
          DbWriteFile (file_name, text);
          BuildHeaderFile (type);
          BuildCppHeaderFile (type);
          DbTStructureInstall (type);

          {
            CmMessage answer;
            char* text;
            
            answer = CmMessageNew ();

            sprintf (file_name, "%s.h", DbTypeName (type));
            text = DbReadFile (file_name);
            CmMessagePutText (answer, file_name);
            CmMessagePutText (answer, text);
            free (text);

            sprintf (file_name, "%s.hxx", DbTypeName (type));
            text = DbReadFile (file_name);
            CmMessagePutText (answer, file_name);
            CmMessagePutText (answer, text);
            free (text);

            CmMessageSetType (answer, "DbTypeHeaders");
            CmMessageSend (answer, sender);

            CmMessageDelete (answer);
          }
        }
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "CreateTypeHandler>");
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus LoadTypeHandler (CmMessage message, 
                                        char* sender,
                                        char* server)
/*--------------------------------------------------------------------*/
{
  char* name;
  int transaction = 0;
  
  name = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("LoadTypeHandler> from %s name = %s\n", sender, name);
    }
  
  if (name != NULL)
    {
      DbType type;
      
      DbClearErrorInfo ();
      
      type = DbTStructureFind (name);
      
      if (type == NULL)
        {
          type = LoadType (name);
        }
      
      if (type != NULL)
        {
          CmMessage answer;
          char* text;
          
          answer = CmMessageNew ();
          text = FormatType (type);
          CmMessagePutText (answer, text);
          CmMessageSetType (answer, "DbTypeDefinition");
          CmMessageSend (answer, sender);
          CmMessageDelete (answer);
        }
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "LoadTypeHandler>");
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}


/*--------------------------------------------------------------------*/
static DbObject CreateObjectAction (char* text)
/*--------------------------------------------------------------------*/
{
  DbObject object = NULL;
      
  if (DbDebug ())
    {
      printf ("CreateObjectAction> text = %s\n", text);
    }
    
  /*GCReferenceAssign (object, DbParserCreateObject (text));*/
  object = DbParserCreateObject (text);

  if (object == NULL)
    {
        /*
          There was an error while defining the object.
          Typically the possible causes can be :
          - a syntax error
          - a bad redefinition of an existing object
          */
    }
  else
    {
        /*
          The object is correctly defined. If this is a
          redefinition of an existing object, the most we
          may do is to change the value - not the type
          */
      
      char* file_name;
      char* name;
      char* path_name;
      DbObjectPath path;
      
      name = DbObjectGetName (object);
      path_name = DbObjectGetPathName (object);
      
      if (DbObjectFindInPath (name, path_name) != NULL)
        {
          DbSetErrorInfo (DbErrorObjectAlreadyDefined,
                          "CreateObjectAction> %s in path %s", 
                          name, path_name);
        }
      else
        {
          if (strlen (path_name) > 0)
            {
              char command[256];
              
              strcpy (command, "mkdir -p ");
              strcat (command, path_name);
              system (command);
            }
          
          file_name = BuildFileName (name, path_name);
          
            /*
              Il faudrait plutot reformatter le texte 
              de l'objet
              */
          DbWriteFile (file_name, text);
          
            /* This is a first installation of this object */
          /*
          if (DbObjectInstallInPath (object, path_name))
            {
              DbType type;
              
              type = DbObjectGetType (object);
              type = DbTypeGetUsedStructure (type);
              
              if (type != NULL)
                {
                  DbTStructure structure = (DbTStructure) type;

                  DbTStructureUpdateClients (structure);
                }
            }
            */
        }
    }

  return (object);
}





/*--------------------------------------------------------------------*/
static CmMessageStatus CreateObjectHandler (CmMessage message, 
                                            char* sender,
                                            char* server)
/*--------------------------------------------------------------------*/
{
  char* text;
  int transaction = 0;
  
  DbClearErrorInfo ();
      
  text = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (text != NULL)
    {
      DbObject object = NULL;
      char* name;
      char* path_name;

      GCReferenceAssign (object, CreateObjectAction (text));

      if (object != NULL)
        {
          DbVersioningOption versioning;

          versioning = DbObjectGetVersioningOption (object);
          if (versioning == DbDefaultVersioning) 
            {
              versioning = DefaultVersioningOption;
            }

          if (versioning == DbUseVersioning)
            {
              name = DbObjectGetName (object);
              path_name = DbObjectGetPathName (object);

              DbCommitObjectVersion (path_name, name);
            }

	  DbObjectDestroy (object);
          GCReferenceClear (object);
        }
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "CreateObjectHandler>");
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus CreateObjectsHandler (CmMessage message, 
                                             char* sender,
                                             char* server)
/*--------------------------------------------------------------------*/
{
  char* text;
  int transaction = 0;

  DbClearErrorInfo ();
      
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  while (CmMessageGetItemType (message) == CmMessageItemText)
    {
      text = CmMessageGetText (message);
  
      if (text != NULL)
        {
          DbObject object = NULL;
          char* name;
          char* path_name;

          GCReferenceAssign (object, CreateObjectAction (text));
          
          if (object != NULL)
            {
              DbVersioningOption versioning;
              
              versioning = DbObjectGetVersioningOption (object);
              if (versioning == DbDefaultVersioning) 
                {
                  versioning = DefaultVersioningOption;
                }

              if (versioning == DbUseVersioning)
                {
                  name = DbObjectGetName (object);
                  path_name = DbObjectGetPathName (object);

                  DbCommitObjectVersion (path_name, name);
                }

	      DbObjectDestroy (object);
              GCReferenceClear (object);
            }
          else
            {
                /*
                  The last object creation failed...
                  */
              break;
            }
        }
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus LoadObjectHandler (CmMessage message, 
                                          char* sender,
                                          char* server)
/*--------------------------------------------------------------------*/
{
  char* name;
  char* path_name;
  int transaction = 0;

  name      = CmMessageGetText (message);
  path_name = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("LoadObjectHandler> from %s name = ", sender);
      
      if (path_name == NULL) path_name = "";
      
      if (strlen (path_name) > 0)
        {
          printf ("%s/", path_name);
        }
    
      printf ("%s\n", name);
    }
  
  if ((name != NULL) && (strlen (name) > 0))
    {
      DbObject object = NULL;
      char* full_name = NULL;
      DbLock lock;
      DbClient client;
      
      DbClearErrorInfo ();

      full_name = BuildFullName (name, path_name);

      lock = DbFindLock (full_name);
      client = DbFindClient (sender);

        /*
          Objects are always at least readable but we have to
          signal when a ReadWrite lock is set AND the requested
          object is already locked by another client.
          */

      
      if (DbClientRequestWriteAccess (client) &&
          !DbLockPermitsReadForWrite (lock, sender))
        {
          if (lock == NULL)
            {
              DbSetErrorInfo (DbErrorWriteAccessDenied, 
                              "LoadObjectHandler> not locked");
            }
          else
            {
              DbSetErrorInfo (DbErrorWriteAccessDenied, 
                              "LoadObjectHandler> locked by %s", 
                              DbLockGetOwner (lock));
            }
        }
      else
        {
	  GCReferenceAssign (object, DbObjectFindInPath (name, path_name));
          
          if (object == NULL)
            {
	      GCReferenceAssign (object,
				 DbLoadObjectFromFile (name, path_name));
	      
	      /*
		Installer l'objet
	      */
	    }
        }
      
      
 
      if (object != NULL)
        {
          CmMessage answer;
          char* text;
	  
          if (DbClientRequestWriteAccess (client))
            {
	      
	      /*
		When the sender is declared as a client, AND that
		it has set ReadWrite locks, AND that this point has
		been reached, this implies that a lock must be
		set on this object.
	      */
	      DbLockNew (full_name, client);
            }
          
          answer = CmMessageNew ();
	  text   = DbConvertObjectToString (object);
          CmMessagePutText (answer, text);
          CmMessageSetType (answer, "DbObjectDefinition");
          CmMessageSend    (answer, sender);
          CmMessageDelete (answer);

	  DbObjectDestroy (object);
          GCReferenceClear (object);
	}
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "LoadObjectHandler>");
    }
  
  DbSendBreak (sender, transaction);

  GCCleanup ();
  /*  GCDumpBase ();*/
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus LoadObjectsHandler (CmMessage message,
                                           char* sender, 
                                           char* server)
/*---------------------------------------------------------------------*/
{
  char*  text;

  char*  path_reg;
  char*  name_reg;
  int    transaction = 0;
  char   search_command[512];
  FILE*  command;
  
  if (DbDebug ())
    {
      printf ("LoadObjectsHandler> from %s\n", sender);
    }
  
  DbClearErrorInfo ();
  
  name_reg = CmMessageGetText (message);
  path_reg = CmMessageGetText (message);
  
  if (DbDebug ())
    {
      printf ("LoadObjectsHandler> path : %s - objects : %s\n", 
              path_reg, name_reg);
    }
      
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  strcpy (search_command, "ls ");
  strcat (search_command, BuildFileName (name_reg, path_reg));

  command = popen (search_command, "r");
  if (command != NULL)
    {
      CmMessage answer;
      char line[256];

      answer = CmMessageNew ();
      CmMessagePutInt  (answer, transaction);

      while (fgets (line, sizeof (line), command) != NULL)
        {
          char* name;
          char* path_name;
          char* full_name;
          char* slash;
          char* nl;
          char* suffix;
          DbObject object = NULL;
          DbLock lock;
          DbClient client;
      
          nl = strchr (line, '\n');
          if (nl != NULL) *nl = 0;

          suffix = strstr (line, ".db");
          if (suffix != NULL) *suffix = 0;

          slash = strrchr (line, '/');
          if (slash == NULL)
            {
              name = line;
              path_name = "";
            }
          else
            {
              name = slash + 1;
              *slash = 0;
              path_name = line;
            }

          full_name = BuildFullName (name, path_name);

          lock = DbFindLock (full_name);
          client = DbFindClient (sender);

            /*
              Objects are always at least readable but we have to
              signal when a ReadWrite lock is set AND the requested
              object is already locked by another client.
              */

          if (DbClientRequestWriteAccess (client) &&
              !DbLockPermitsReadForWrite (lock, sender))
            {
              if (lock == NULL)
                {
                  DbSetErrorInfo (DbErrorWriteAccessDenied, 
                                  "LoadObjectsHandler> not locked");
                }
              else
                {
                  DbSetErrorInfo (DbErrorWriteAccessDenied, 
                                  "LoadObjectsHandler> locked by %s", 
                                  DbLockGetOwner (lock));
                }
                /*
                  The search is stopped at the first access error.
                  */
              break;
            }
          else
            {
              GCReferenceAssign (object,
                                 DbObjectFindInPath (name, path_name));
          
              if (object == NULL)
                {
                  GCReferenceAssign (object,
                                     DbLoadObjectFromFile (name, path_name));
                    /*
                      Installer l'objet
                      */
                }
            }

          if (object != NULL)
            {
              char* text;
              
              if (DbClientRequestWriteAccess (client))
                {
                    /*
                      When the sender is declared as a client, AND that
                      it has set ReadWrite locks, AND that this point has
                      been reached, this implies that a lock must be
                      set on this object.
                      */
                  DbLockNew (full_name, client);
                }
              
              text = DbConvertObjectToString (object);
              CmMessagePutText (answer, text);

              DbObjectDestroy (object);
              GCReferenceClear (object);
            }
        }
      
      pclose (command);

      CmMessageSetType (answer, "DbObjectDefs");
      CmMessageSend    (answer, sender);
      CmMessageDelete  (answer);
    }

  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}




/*--------------------------------------------------------------------*/
static void CommitObject (DbObject object)
/*--------------------------------------------------------------------*/
{
  char*  name;
  char*  path_name;
  char*  file_name;

  char*  text;  

  if (object == NULL) return;
  
  name      = DbObjectGetName (object);
  path_name = DbObjectGetPathName (object);

  file_name = BuildFileName (name, path_name);
  text      = DbConvertObjectToString (object);
  
  DbWriteFile (file_name, text);
}

/*--------------------------------------------------------------------*/
static void CommitObjectVersion (DbObject object)
/*--------------------------------------------------------------------*/
{
  char*  name;
  char*  path_name;
  char*  file_name;

  char*  text;  

  DbVersioningOption versioning;

  if (object == NULL) return;
  
  name      = DbObjectGetName (object);
  path_name = DbObjectGetPathName (object);

  versioning = DbObjectGetVersioningOption (object);
  if (versioning == DbDefaultVersioning)
    versioning = DefaultVersioningOption;
  
  if (versioning == DbUseVersioning)
    DbCommitObjectVersion (path_name, name);

}

/*--------------------------------------------------------------------*/
static CmMessageStatus CommitObjectHandler (CmMessage message, 
                                            char* sender,
                                            char* server)
/*--------------------------------------------------------------------*/
{
  char* text;
  int transaction = 0;
  DbObject object = NULL;
  char full_name[256];
  DbLock my_lock = NULL;
  static DbClient myself = NULL;
        
  if (myself == NULL)
    {
      GCReferenceAssign (myself, DbClientNew (DbServerName ()));
    }

  text = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
  {
    transaction = CmMessageGetInt (message);
  }

  if (DbDebug ())
  {
    printf ("CommitObjectHandler> from %s text = [%s]\n", sender, text);
  }
  
  if (text != NULL)
  {
    DbClearErrorInfo ();
      
    GCReferenceAssign (object, DbParserCreateObject (text));
      
    if (object == NULL)
    {
          /*
            There was an error while defining the object.
            Typically the possible causes can be :
            - a syntax error
            - a bad redefinition of an existing object
            */
    }
    else
    {
      char*  name;
      char*  path_name;
      
      DbLock lock;
              /*
                The object is correctly defined. If this is a
                redefinition of an existing object, the most we
                may do is to change the value - not the type
                */
      name      = DbObjectGetName (object);
      path_name = DbObjectGetPathName (object);
      strcpy (full_name, BuildFullName (name, path_name));
          
      lock = DbFindLock (full_name);
      if (!DbLockPermitsWrite (lock, sender))
      {
        if (lock == NULL)
        {
          DbSetErrorInfo (DbErrorWriteAccessDenied, 
                          "CommitObjectHandler> not locked");
        }
        else
        {
          DbSetErrorInfo (DbErrorWriteAccessDenied, 
                          "CommitObjectHandler> locked by %s", 
                          DbLockGetOwner (lock));
        }
      }
      else
      {
          /*
            On memorise le lock actuel, et DbServer depose un lock
            pour lui-meme
            */
        DbLockRemove (lock);
        my_lock = DbLockNew (full_name, myself);
            
        CommitObject (object);
        /*
          Le commit CVS est reporte apres l'envoi du break...
          A verifier...
          */
      }
    }
  }
  else
  {
    DbSetErrorInfo (DbErrorBadArgument, "CommitObjectHandler>");
  }
  
  DbSendBreak (sender, transaction);

  if (object != NULL)
  {
    /*
      Attention aux entrelacements de commits car le break
      precedent peut avoir recu un autre commit sur le meme
      objet !!!
      */
    CommitObjectVersion (object);

    if (my_lock != NULL)
      {
        DbLockRemove (my_lock);
        DbSendTriggers (full_name, DbTriggerUnlock);
      }

    DbObjectDestroy (object);
    GCReferenceClear (object);

    DbSendTriggers (full_name, DbTriggerCommit);
  }

  /*   No cleanup to speed up the commit. Note that it is necessary */
  /*  GCCleanup ();*/


  /*  GCDumpBase ();*/
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus CommitObjectsHandler (CmMessage message, 
                                             char* sender,
                                             char* server)
/*--------------------------------------------------------------------*/
{
  int transaction = 0;
  CLList objects = NULL;
  static DbClient myself = NULL;
  
  if (DbDebug ())
    {
      printf ("CommitObjectsHandler> from %s\n", sender);
    }
  
  if (myself == NULL)
    {
      GCReferenceAssign (myself, DbClientNew (DbServerName ()));
    }

  DbClearErrorInfo ();

  while (CmMessageGetItemType (message) == CmMessageItemText)
  {
    char* text;
    
    text = CmMessageGetText (message);
    
    if (text != NULL)
    {
      DbObject object = NULL;
      
      DbClearErrorInfo ();
      
      GCReferenceAssign (object, DbParserCreateObject (text));

      if (object == NULL)
      {
            /*
              There was an error while defining the object.
              Typically the possible causes can be :
              - a syntax error
              - a bad redefinition of an existing object
              */
      }
      else
      {
        char*  name;
        char*  path_name;
        char*  full_name;
        
        DbLock lock;
            /*
              The object is correctly defined. If this is a
              redefinition of an existing object, the most we
              may do is to change the value - not the type
              */
        if (objects == NULL)
        {
          GCReferenceAssign (objects, CLListNew ());
        }
        CLListAdd (objects, object);
      
        name      = DbObjectGetName (object);
        path_name = DbObjectGetPathName (object);
        full_name = BuildFullName (name, path_name);
        
        lock = DbFindLock (full_name);
        if (!DbLockPermitsWrite (lock, sender))
        {
          if (lock == NULL)
          {
            DbSetErrorInfo (DbErrorWriteAccessDenied, 
                            "CommitObjectsHandler> not locked");
          }
          else
          {
            DbSetErrorInfo (DbErrorWriteAccessDenied, 
                            "CommitObjectsHandler> locked by %s", 
                            DbLockGetOwner (lock));
          }
        }
        else
        {
          DbLockRemove (lock);
          DbLockNew (full_name, myself);

          CommitObject (object);
        }

	DbObjectDestroy (object);
        GCReferenceClear (object);
      }
    }
    else
    {
      DbSetErrorInfo (DbErrorBadArgument, "CommitObjectsHandler>");
    }
  }

  if (CmMessageGetItemType (message) == CmMessageItemInt)
  {
    transaction = CmMessageGetInt (message);
  }

  DbSendBreak (sender, transaction);

  if (objects != NULL)
  {
    CIterator it;

    it = CLListGiveIterator (objects);
    for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbObject object;
      char*  name;
      char*  path_name;
      char*  full_name;
      DbLock lock;

      object = (DbObject) CIteratorItem (it);
      CommitObjectVersion (object);

      name      = DbObjectGetName (object);
      path_name = DbObjectGetPathName (object);
      full_name = BuildFullName (name, path_name);

      lock = DbFindLock (full_name);
      if (lock != NULL) 
        {
          DbLockRemove (lock);          
          DbSendTriggers (full_name, DbTriggerUnlock);
        }

      DbSendTriggers (full_name, DbTriggerCommit);
    }

    CIteratorDelete (it);

    DbObjectsDestroy (objects);
    CLListErase (objects);
    GCReferenceClear (objects);
  }
  
  return (CmMessageOk);
}



/*--------------------------------------------------------------------*/
static CmMessageStatus CommitObjectAndHoldLockHandler (CmMessage message, 
                                                       char* sender,
                                                       char* server)
/*--------------------------------------------------------------------*/
{
  char* text;
  int transaction = 0;
  DbObject object = NULL;
  char full_name[256];
  DbLock my_lock = NULL;
  static DbClient myself = NULL;
  
  if (myself == NULL)
    {
      GCReferenceAssign (myself, DbClientNew (DbServerName ()));
    }

  text = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("CommitObjectAndHoldLockHandler> from %s text = [%s]\n", 
              sender, text);
    }
  
  if (text != NULL)
    {
      DbClearErrorInfo ();
      
      GCReferenceAssign (object, DbParserCreateObject (text));
      
      if (object == NULL)
        {
            /*
              There was an error while defining the object.
              Typically the possible causes can be :
              - a syntax error
              - a bad redefinition of an existing object
              */
        }
      else
        {
            /*
              The object is correctly defined. If this is a
              redefinition of an existing object, the most we
              may do is to change the value - not the type
              */
          
          char* name;
          char* path_name;
          DbLock lock;

          DbVersioningOption versioning;
          
          name = DbObjectGetName (object);
          path_name = DbObjectGetPathName (object);
          strcpy (full_name, BuildFullName (name, path_name));

          lock = DbFindLock (full_name);
          if (!DbLockPermitsWrite (lock, sender))
            {
              if (lock == NULL)
                {
                  DbSetErrorInfo (DbErrorWriteAccessDenied, 
                                  "CommitObjectAndHoldLockHandler> not locked");
                }
              else
                {
                  DbSetErrorInfo (DbErrorWriteAccessDenied, 
                                  "CommitObjectAndHoldLockHandler> "
                                  "locked by %s", 
                                  DbLockGetOwner (lock));
                }
            }
          else
            {
              /*
                On memorise le lock actuel, et DbServer depose un lock
                pour lui-meme
                */
              DbLockRemove (lock);
              my_lock = DbLockNew (full_name, myself);

              CommitObject (object);

                /*
                  Le commit CVS est reporte apres l'envoi du break...
                  A verifier...
                  */
            }
        }
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "CommitObjectAndHoldLockHandler>");
    }
  
  DbSendBreak (sender, transaction);

  if (object != NULL)
  {
    /*
      Attention aux entrelacements de commits car le break
      precedent peut avoir recu un autre commit sur le meme
      objet !!!
      -> il faudrait installer un lock pour DbServer lui-meme avec
      restauration du lock d'origine lorsque tout est termine.
      */
    CommitObjectVersion (object);

    if (my_lock != NULL)
      {
        DbLockRemove (my_lock);
        DbLockNew (full_name, DbClientNew (sender));
      }

    GCReferenceClear (object);

    DbSendTriggers (full_name, DbTriggerCommit);  
  }

  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus LockObjectHandler (CmMessage message, 
                                          char* sender,
                                          char* server)
/*--------------------------------------------------------------------*/
{
  char* name;
  char* path_name;
  int transaction = 0;
  
  name      = CmMessageGetText (message);
  path_name = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("LockObjectHandler> from %s name = ", sender);
      
      if (path_name == NULL) path_name = "";
      
      if (strlen (path_name) > 0)
        {
          printf ("%s/", path_name);
        }
    
      printf ("%s\n", name);
    }
  
  if ((name != NULL) && (strlen (name) > 0))
    {
      char* full_name = NULL;
      DbLock lock;
      DbClient client;
      
      DbClearErrorInfo ();

      full_name = BuildFullName (name, path_name);

      lock = DbFindLock (full_name);
      client = DbFindClient (sender);
      if (client == NULL)
        {
          client = DbClientNew (sender);
        }

        /*
          Objects are always at least readable but we have to
          signal when a ReadWrite lock is set AND the requested
          object is already locked by another client.
          */

      if (!DbLockPermitsReadForWrite (lock, sender))
        {
          if (lock == NULL)
            {
              DbSetErrorInfo (DbErrorWriteAccessDenied, 
                              "LockObjectHandler> not locked");
            }
          else
            {
              DbSetErrorInfo (DbErrorWriteAccessDenied, 
                              "LockObjectHandler> locked by %s", 
                              DbLockGetOwner (lock));
            }
        }
      else
        {
            /*
              When the sender is declared as a client, AND that
              it has set ReadWrite locks, AND that this point has
              been reached, this implies that a lock must be
              set on this object.
              */
          DbLockNew (full_name, client);
        }
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "LockObjectHandler>");
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus UnlockObjectHandler (CmMessage message, 
                                            char* sender,
                                            char* server)
/*--------------------------------------------------------------------*/
{
  char* name;
  char* path_name;
  int transaction = 0;
  
  name      = CmMessageGetText (message);
  path_name = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("UnlockObjectHandler> from %s name = ", sender);
      
      if (path_name == NULL) path_name = "";
      
      if (strlen (path_name) > 0)
        {
          printf ("%s/", path_name);
        }
    
      printf ("%s\n", name);
    }
  
  if ((name != NULL) && (strlen (name) > 0))
    {
      char* full_name = NULL;
      DbLock lock;
      DbClient client;
      
      DbClearErrorInfo ();

      full_name = BuildFullName (name, path_name);

      lock = DbFindLock (full_name);

      if (lock != NULL)
        {
          DbLockRemove (lock);
        }

      DbSendTriggers (full_name, DbTriggerUnlock);
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "UnlockObjectHandler>");
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus AddTriggerHandler (CmMessage message, 
                                          char* sender,
                                          char* server)
/*--------------------------------------------------------------------*/
{
  char* name;
  char* path_name;
  int transaction = 0;
  int trigger_value;
  DbTriggerType trigger_type;

  name      = CmMessageGetText (message);
  path_name = CmMessageGetText (message);
  
  trigger_value = CmMessageGetInt (message);
  if (DbDebug ())
    printf ("AddTriggerHandler ()  trigger = %d\n", trigger_value);

  switch (trigger_value)
    {
      case 1 :
        trigger_type = DbTriggerCommit;
        break;
      case 2 :
        trigger_type = DbTriggerUnlock;
        break;
      default :
        trigger_type = DbTriggerNone;
        break;
    }

  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("AddTriggerHandler> from %s name = ", sender);
      
      if (path_name == NULL) path_name = "";
      
      if (strlen (path_name) > 0)
        {
          printf ("%s/", path_name);
        }
    
      printf ("%s\n", name);
    }
  
  if ((name != NULL) && (strlen (name) > 0))
    {
      char* full_name = NULL;
      DbTrigger trigger;
      DbClient client;
      
      DbClearErrorInfo ();

      full_name = BuildFullName (name, path_name);
      client = DbClientNew (sender);

      DbTriggerAdd (full_name, trigger_type, client);
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "AddTriggerHandler>");
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus RemoveTriggerHandler (CmMessage message, 
                                             char* sender,
                                             char* server)
/*--------------------------------------------------------------------*/
{
  char* name;
  char* path_name;
  int transaction = 0;
  int trigger_value;
  DbTriggerType trigger_type;

  name      = CmMessageGetText (message);
  path_name = CmMessageGetText (message);
  
  trigger_value = CmMessageGetInt (message);
  if (DbDebug ())
    printf ("RemoveTriggerHandler ()  trigger = %d\n", trigger_value);

  switch (trigger_value)
    {
      case 1 :
        trigger_type = DbTriggerCommit;
        break;
      case 2 :
        trigger_type = DbTriggerUnlock;
        break;
      default :
        trigger_type = DbTriggerNone;
        break;
    }

  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("RemoveTriggerHandler> from %s name = ", sender);
      
      if (path_name == NULL) path_name = "";
      
      if (strlen (path_name) > 0)
        {
          printf ("%s/", path_name);
        }
    
      printf ("%s\n", name);
    }
  
  if ((name != NULL) && (strlen (name) > 0))
    {
      char* full_name = NULL;
      DbTrigger trigger;
      DbClient client;
      
      DbClearErrorInfo ();

      full_name = BuildFullName (name, path_name);
      client = DbClientNew (sender);

      DbTriggerRemove (full_name, trigger_type, client);
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "RemoveTriggerHandler>");
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus DeleteObjectHandler (CmMessage message, 
                                            char* sender,
                                            char* server)
/*--------------------------------------------------------------------*/
{
  char* name;
  char* path_name;
  int transaction = 0;
  
  name      = CmMessageGetText (message);
  path_name = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("DeleteObjectHandler> from %s name = %s\n", sender, name);
    }
  
  if (name != NULL)
    {
      DbObject object = NULL;
      
      DbClearErrorInfo ();
      
      object = DbObjectFindInPath (name, path_name);
      
      if (object == NULL)
        {
          object = DbLoadObjectFromFile (name, path_name);
            /*
              Installer l'objet
              */
        }
      
      if (object != NULL)
        {
          DbType type;
          char* file_name;
          DbVersioningOption versioning;
          
          type = DbObjectGetType (object);
          type = DbTypeGetUsedStructure (type);
          if (type != NULL)
            {
              DbTStructure structure = (DbTStructure) type;

              DbTStructureUpdateClients (structure);
            }

          versioning = DbObjectGetVersioningOption (object);
          if (versioning == DbDefaultVersioning)
            versioning = DefaultVersioningOption;

          DbObjectRemove (object);
          
          file_name = BuildFileName (name, path_name);
          unlink (file_name);
          
          if (versioning == DbUseVersioning)
            DbRemoveObjectVersion (path_name, name);      
        }
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "DeleteObjectHandler>");
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus DeleteTypeHandler (CmMessage message, 
                                          char* sender,
                                          char* server)
/*--------------------------------------------------------------------*/
{
  char* name;
  int transaction = 0;
  
  name = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("DeleteTypeHandler> from %s name = %s\n", sender, name);
    }
  
  if (name != NULL)
    {
      DbType type;
      char file_name[256];
      
      DbClearErrorInfo ();
      
      type = DbTStructureFind (name);
      
      if (type != NULL)
        {
          DbTStructureUninstall (type);
        }

      sprintf (file_name, "%s.ddl", name);
      unlink (file_name);
      sprintf (file_name, "%s.h", name);
      unlink (file_name);
      sprintf (file_name, "%s.hxx", name);
      unlink (file_name);
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "DeleteTypeHandler>");
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus UpdateTypeHandler (CmMessage message, 
                                          char* sender,
                                          char* server)
/*--------------------------------------------------------------------*/
{
  char* name;
  int transaction = 0;
  
  name = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("UpdateTypeHandler> from %s name = %s\n", sender, name);
    }
  
  if (name != NULL)
    {
      DbType type;
      
      DbClearErrorInfo ();
      
      type = DbTStructureFind (name);
      
      if (type == NULL)
        {
          type = LoadType (name);
        }
      
      if (type != NULL) 
        {
          BuildHeaderFile (type);
          BuildCppHeaderFile (type);
        }

      {
        CmMessage answer;
        char* text;
        char file_name[256];
            
        answer = CmMessageNew ();
        
        sprintf (file_name, "%s.h", DbTypeName (type));
        text = DbReadFile (file_name);
        CmMessagePutText (answer, file_name);
        CmMessagePutText (answer, text);
        free (text);
        
        sprintf (file_name, "%s.hxx", DbTypeName (type));
        text = DbReadFile (file_name);
        CmMessagePutText (answer, file_name);
        CmMessagePutText (answer, text);
        free (text);
        
        CmMessageSetType (answer, "DbTypeHeaders");
        CmMessageSend (answer, sender);
        
        CmMessageDelete (answer);
      }
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "UpdateTypeHandler>");
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus UpdateHeadersHandler (CmMessage message, 
                                             char* sender,
                                             char* server)
/*--------------------------------------------------------------------*/
{
  DbType type;
  char* command = "ls *.ddl";
  char line[256];
  FILE* list;
  int transaction = 0;
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("UpdateHeadersHandler> from %s\n", sender);
    }
  
  DbClearErrorInfo ();
  
  list = popen (command, "r");
  if (list != NULL)
    {
      while (!feof (list))
        {
          char* ptr;
          char* nl;
          char* dot;
          
          line[0] = 0;
          
          ptr = fgets (line, sizeof (line), list);
          if (ptr == NULL) break;
          
          dot = strrchr (line, '.');
          if (dot != NULL)
            {
              *dot = 0;
              
              type = DbTStructureFind (line);
              
              if (type == NULL)
                {
                  type = LoadType (line);
                }
              
              if (type != NULL) 
                {
                  BuildHeaderFile (type);
                  BuildCppHeaderFile (type);
                }
            }
        }
      
      pclose (list);
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus ObjectGetNamesHandler (CmMessage message, 
                                              char* sender,
                                              char* server)
/*--------------------------------------------------------------------*/
{
  char* expression;
  int transaction = 0;
  
  expression = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("ObjectGetNamesHandler> from %s expression = %s\n", 
              sender, expression);
    }
  
  if (expression != NULL)
    {
      CmMessage answer = NULL;
      char command[256];
      char line[256];
      FILE* list;
      char* slash;
      
      answer = CmMessageNew ();
      
      DbClearErrorInfo ();
      
      slash = strrchr (expression, '/');
      
      if (slash != NULL)
        {
          sprintf (command, "ls %s.db", expression);
        }
      else
        {
          sprintf (command, "find . -name '%s.db' | sed 's/^[.][/]//'", 
                   expression);
        }
      
      list = popen (command, "r");
      if (list != NULL)
        {
          while (!feof (list))
            {
              char* ptr;
              char* nl;
              char* dot;
              
              line[0] = 0;
              
              ptr = fgets (line, sizeof (line), list);
              if (ptr == NULL) break;
              nl = strchr (line, '\n');
              if (nl != NULL) *nl = 0;
              
              dot = strrchr (line, '.');
              if (dot != NULL)
                {
                  *dot = 0;
                  
                  CmMessagePutText (answer, line);
                }
            }
              
          pclose (list);
        }
      
      CmMessageSetType (answer, "DbObjectNames");
      CmMessageSend (answer, sender);
      CmMessageDelete (answer);
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "ObjectNamesHandler>");
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus ObjectGetNamesInPathHandler (CmMessage message, 
                                                    char* sender,
                                                    char* server)
/*--------------------------------------------------------------------*/
{
  char* path;
  CmMessage answer = NULL;
  char command[256];
  char line[256];
  FILE* list;
  int transaction = 0;
  
  path = CmMessageGetText (message);
  if ((path == NULL) || (strlen (path) == 0)) path = ".";
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("ObjectGetNamesInPathHandler> from %s path = %s\n", 
              sender, path);
    }
  
  answer = CmMessageNew ();
  
  DbClearErrorInfo ();
  
  sprintf (command, "(cd %s; ls *.db)", path);
  
  list = popen (command, "r");
  if (list != NULL)
    {
      while (!feof (list))
        {
          char* ptr;
          char* nl;
          char* dot;
          
          line[0] = 0;
          
          ptr = fgets (line, sizeof (line), list);
          if (ptr == NULL) break;
          nl = strchr (line, '\n');
          if (nl != NULL) *nl = 0;
          
          dot = strrchr (line, '.');
          if (dot != NULL)
            {
              *dot = 0;
              
              CmMessagePutText (answer, line);
            }
        }
      
      pclose (list);
    }
  
  CmMessageSetType (answer, "DbObjectNames");
  CmMessageSend (answer, sender);
  CmMessageDelete (answer);
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus SubPathGetNamesHandler (CmMessage message, 
                                               char* sender,
                                               char* server)
/*--------------------------------------------------------------------*/
{
  char* path;
  CmMessage answer = NULL;
  char command[256];
  char line[256];
  FILE* list;
  int transaction = 0;
  
  path = CmMessageGetText (message);
  if ((path == NULL) || (strlen (path) == 0)) path = ".";
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("SubPathGetNamesHandler> from %s path = %s\n", 
              sender, path);
    }
  
  answer = CmMessageNew ();
  
  DbClearErrorInfo ();
  
  sprintf (command, "(cd %s; ls | grep -v '\\.')", path);
  
  list = popen (command, "r");
  if (list != NULL)
    {
      while (!feof (list))
        {
          char* ptr;
          char* nl;
          
          line[0] = 0;
          
          ptr = fgets (line, sizeof (line), list);
          if (ptr == NULL) break;
          nl = strchr (line, '\n');
          if (nl != NULL) *nl = 0;
          if (strcmp (line, "CVS") == 0) continue;
          
          CmMessagePutText (answer, line);
        }
      
      pclose (list);
    }
  
  CmMessageSetType (answer, "DbObjectNames");
  CmMessageSend (answer, sender);
  CmMessageDelete (answer);
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/
static CmMessageStatus TypeGetNamesHandler (CmMessage message, 
                                            char* sender,
                                            char* server)
/*--------------------------------------------------------------------*/
{
  char* expression;
  int transaction = 0;
  
  expression = CmMessageGetText (message);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  if (DbDebug ())
    {
      printf ("GetTypeNamesHandler> from %s expression = %s\n", 
              sender, expression);
    }
  
  if (expression != NULL)
    {
      CmMessage answer = NULL;
      char command[256];
      char line[256];
      FILE* list;
      
      answer = CmMessageNew ();
      
      DbClearErrorInfo ();
      
      sprintf (command, "ls %s.ddl", expression);
      
      list = popen (command, "r");
      if (list != NULL)
        {
          while (!feof (list))
            {
              char* ptr;
              char* nl;
              char* dot;
              
              line[0] = 0;
              
              ptr = fgets (line, sizeof (line), list);
              if (ptr == NULL) break;
              
              dot = strrchr (line, '.');
              if (dot != NULL)
                {
                  *dot = 0;
                  
                  CmMessagePutText (answer, line);
                }
            }
          
          pclose (list);
        }
      
      CmMessageSetType (answer, "DbTypeNames");
      CmMessageSend (answer, sender);
      CmMessageDelete (answer);
    }
  else
    {
      DbSetErrorInfo (DbErrorBadArgument, "TypeNamesHandler>");
    }
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------*/
static CmMessageStatus CommandHandler (CmMessage message, 
                                       char* sender,
                                       char* server)
/*--------------------------------------------------------*/
{
  char* command;
  char* space;
  int transaction = 0;
  
  command = CmMessageGetText (message);
  if (DbDebug ())
    printf ("CommandHandler ()  command = %s\n", command);
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  DbClearErrorInfo ();
  
  DbSendBreak (sender, transaction);
  
  space = strchr (command, ' ');
  if (space) *space = 0;
  
  if (!strcmp (command, "Restart"))
    {
      Finished = 2;
      CmTerminateTransaction (Transaction);
    }
  
  if (!strcmp (command, "Stop"))
    {
      Finished = 1;
      CmTerminateTransaction (Transaction);
    }
  
  return (CmMessageOk);
}

/*--------------------------------------------------------*/
static CmMessageStatus SetLocksHandler (CmMessage message, 
                                        char* sender,
                                        char* server)
/*--------------------------------------------------------*/
{
  int mode_value = 0;
  int transaction = 0;
  DbClient client;
  DbLockMode mode;
  
  mode_value = CmMessageGetInt (message);
  if (DbDebug ())
    printf ("SetLocksHandler ()  mode = %d\n", mode_value);

  switch (mode_value)
    {
      case 1 :
        mode = DbLockReadWrite;
        break;
      default :
        mode = DbLockReadOnly;
        break;
    }

  client = DbClientNew (sender);
  DbClientSetLockMode (client, mode);

  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  DbClearErrorInfo ();
  
  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------*/
static CmMessageStatus ResetLocksHandler (CmMessage message, 
                                          char* sender,
                                          char* server)
/*--------------------------------------------------------*/
{
  int transaction = 0;
  DbClient client;

  if (DbDebug ())
    printf ("ResetLocksHandler ()\n");
  
  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      transaction = CmMessageGetInt (message);
    }

  DbClearErrorInfo ();
  
  client = DbFindClient (sender);
  if (client != NULL)
    {
      DbClientSetLockMode (client, DbLockReadOnly);
    }

  DbSendBreak (sender, transaction);
  
  return (CmMessageOk);
}

/*--------------------------------------------------------------------*/


/*--------------------------------------------------------------*/
static CmMessageStatus GetVersionHandler (CmMessage message, 
					  char* sender,
					  char* server)
/*--------------------------------------------------------------*/
{
  int status ;

  status = SendVersion (sender);

  if (!status)
    {
      printf ("NSGetVersionHandler> Error sending answer to %s\n",
	      sender);
    }

  return (CmMessageOk);
}

/*--------------------------------------------------------*/
static int SendVersion (char* dest)
/*--------------------------------------------------------*/
{
  CmMessage message = NULL;
  int status;

  message = CmMessageNew ();

  CmMessagePutText (message, DBVERSION);
  CmMessageSetType (message, "DbVersion");

  status = CmMessageSend (message, dest);

  CmMessageDelete (message);

  return (status);
}


/*--------------------------------------------------------------------*/
static Buffer* BufferNew ()
/*--------------------------------------------------------------------*/
{
  Buffer* buffer;
  
  buffer = (Buffer*) malloc (sizeof (Buffer));
  
  BufferInitialize (buffer);
  
  return (buffer);
}

/*--------------------------------------------------------------------*/
static void BufferInitialize (Buffer* buffer)
/*--------------------------------------------------------------------*/
{
  if (buffer == NULL) return;
  
  buffer->string = NULL;
  buffer->allocated = 0;
  buffer->length = 0;
}

/*--------------------------------------------------------------------*/
static void BufferReset (Buffer* buffer)
/*--------------------------------------------------------------------*/
{
  if (buffer == NULL) return;
  
  if (buffer->string != NULL)
    {
      buffer->string[0] = 0;
    }
  
  buffer->length = 0;
}

/*--------------------------------------------------------------------*/
static void BufferAdd (Buffer* buffer, char* text)
/*--------------------------------------------------------------------*/
{
  int length;
  
  if (buffer == NULL) return;
  if (text == NULL) return;
  
  length = strlen (text);
  if (length == 0) return;
  
  while ((buffer->length + length + 1) > buffer->allocated)
    {
      buffer->allocated += 1000;
      
      if (buffer->string == NULL)
        {
          buffer->string = (char*) malloc (buffer->allocated);
          buffer->string[0] = 0;
        }
      else
        {
          buffer->string = (char*) realloc (buffer->string,
                                            buffer->allocated);
        }
    }
  
  strcat (buffer->string, text);
  buffer->length += length;
}

/*- Procedure ---------------------------------------------------------*/
/*                                                                     */
/* Used by a server to inform its client that the transaction is       */
/* now completed (successfully of not)                                 */
/*  The last transaction may have filled the error buffer in case of   */
/* trouble. This error info is then installed in the break message.    */
/*                                                                     */
/*---------------------------------------------------------------------*/
int DbSendBreak (char* name, int transaction)
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  int status;
  DbError error;
  
  message = CmMessageNew ();
  
  error = DbGetError ();
  CmMessagePutInt (message, error);
  if (error != DbErrorOk)
    {
      CmMessagePutText (message, DbGetErrorInfo ());
    }

  CmMessagePutInt (message, transaction);
  CmMessagePutInt (message, transaction);
  
  CmMessageSetType (message, "DbBreak");
  status = CmMessageSend (message, name);

  CmMessageDelete (message);

  GCCleanup ();
  
  return (status);
}

/*---------------------------------------------------------------------*/
static void DisconnectionHandler (CmConnect connect)
/*---------------------------------------------------------------------*/
{
  char* name;
  DbClient client;

  name = CmConnectGetName (connect);

  client = DbFindClient (name);
  if (client != NULL)
    {
      DbClientRemove (client);
    }

  if (PreviousDisconnectionHandler != NULL)
    PreviousDisconnectionHandler (connect);
}
