#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <Db.h>
#include <CmMessage.h>

static int BuildHeaderFile (DbType type);

/*---------------------------------------------------------------------------*/
static void Usage ()
/*---------------------------------------------------------------------------*/
{
  fprintf (stderr, "> DbDDL.exe -c <ddl file>      : create a type\n");
  fprintf (stderr, "            -d <type name>     : delete a type\n");
  fprintf (stderr, "            -delete_object <object name>  : delete an object\n");
  fprintf (stderr, "            -object_names <pattern>       : query object names\n");
  fprintf (stderr, "            -path_names <pattern>         : query path names\n");
  fprintf (stderr, "            -u <type name>     : update header for a type\n");
  fprintf (stderr, "            -update_headers    : update all headers\n");
}

/*---------------------------------------------------------------------------*/
int main (int argc, char* argv[])
/*---------------------------------------------------------------------------*/
{
  char name[256];
  char* text;
  char* answer;
  DbType type;
  CmMessage message;
  int transaction;
  enum
  {
    none,
    create,
    delete,
    delete_object,
    object_names,
    path_names,
    update,
    update_headers
  } option = none;

  if (argc <= 1)
    {
      Usage ();
      return (0);
    }
  argc--;
  argv++;

  if (!CmMessageOpenMultipleServer ("DbDDL"))
    {
      fprintf (stderr, "Cannot install DbDDL as a Cm server.\n");
      return (1);
    }

  DbStartup ();

  name[0] = 0;

  while (argc >= 1)
    {
      char* arg = argv[0];

      if (!strcmp (arg, "-c"))
        {
          option = create;
        }
      else if (!strcmp (arg, "-d"))
        {
          option = delete;
        }
      else if (!strcmp (arg, "-delete_object"))
        {
          option = delete_object;
        }
      else if (!strcmp (arg, "-object_names"))
        {
          option = object_names;
        }
      else if (!strcmp (arg, "-path_names"))
        {
          option = path_names;
        }
      else if (!strcmp (arg, "-u"))
        {
          option = update;
        }
      else if (!strcmp (arg, "-update_headers"))
        {
          option = update_headers;
        }
      else
        {
          strcpy (name, arg);
        }
      argc--;
      argv++;
    }


  message = CmMessageNew ();

  switch (option)
    {
      case none :
        Usage ();
        return (1);
        break;
      case create :
        text = DbReadFile (name);

        if (text == NULL)
          {
            fprintf (stderr, "Cannot open file %s\n", name);
            return (1);
          }
        
        CmMessagePutText (message, text);
        free (text);
        CmMessageSetType (message, "DbCreateType");
        break;
      case delete :
        CmMessagePutText (message, name);
        CmMessageSetType (message, "DbDeleteType");
        break;
      case delete_object :
        DbDeleteObject (name);
        DbPrintError ();
        exit (0);
        break;
      case object_names :
        answer = DbObjectGetNames (name);
        DbPrintError ();
        if (answer != NULL)
          printf ("%s\n", answer);

        exit (0);
        break;
      case path_names :
        answer = DbSubPathGetNames (name);
        DbPrintError ();
        if (answer != NULL)
          printf ("%s\n", answer);

        exit (0);
        break;
      case update :
        CmMessagePutText (message, name);
        CmMessageSetType (message, "DbUpdateType");
        break;
      case update_headers :
        CmMessageSetType (message, "DbUpdateHeaders");
        break;
    }

  transaction = CmOpenTransaction ("DbDDL", NULL);

  CmMessagePutInt (message, transaction);

  if (!CmMessageSend (message, DbServerName ()))
    {
      DbSetErrorInfo (DbErrorNoServer, "Cannot contact the DbServer");
    }
  else
    {
      DbWaitForTransaction (transaction);
    }

  CmCloseTransaction (transaction);

  DbPrintError ();

  return (0);
}


