#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <Db.h>

/*---------------------------------------------------------------------------*/
/*                                                                           */
/*  Default visitor for simple object dump.                                  */
/*                                                                           */
/*---------------------------------------------------------------------------*/

typedef struct _DumperRec* Dumper;
typedef struct _DumperRec
{
  DbVisitorRec base;
  int tab;
} DumperRec;

/*---------------------------------------------------------------------------*/
static void tab (Dumper visitor)
/*---------------------------------------------------------------------------*/
{
  int i;

  for (i = 0; i < visitor->tab; i++)
    {
      printf ("  ");
    }
}

/*---------------------------------------------------------------------------*/
static int down (Dumper visitor, DbTypeIterator it)
/*---------------------------------------------------------------------------*/
{
  char* key;

  if (visitor == NULL) return (0);

  tab (visitor);

  key = DbTypeIteratorGetKey (it);
  if (key != NULL)
    {
      printf ("%s = ", key);
    }

  printf ("{\n");
  visitor->tab++;

  return (1);
}

/*---------------------------------------------------------------------------*/
static int in (Dumper visitor, DbTypeIterator it)
/*---------------------------------------------------------------------------*/
{
  void* address;

  if (visitor == NULL) return (0);

  address = DbTypeIteratorGetItemAddress (it);
  if (address != NULL)
    {
      DbType type;
      char* text;
      
      type    = DbTypeIteratorGetItemType (it);
      
      text = DbTypeOutput (type, address);

      if (text != NULL)
        {
          char* key;

          tab (visitor);

          key = DbTypeIteratorGetKey (it);
          if (key != NULL)
            {
              printf ("%s = ", key);
            }

          printf ("%s\n", text);
        }
    }

  return (1);
}

/*---------------------------------------------------------------------------*/
static int up (Dumper visitor, DbTypeIterator it)
/*---------------------------------------------------------------------------*/
{
  if (visitor == NULL) return (0);

  visitor->tab--;
  tab (visitor);
  printf ("}\n");

  return (1);
}

/*---------------------------------------------------------------------------*/
DbVisitor DumperNew ()
/*---------------------------------------------------------------------------*/
{
  Dumper object;

  object = (Dumper) malloc (sizeof (DumperRec));

  DbVisitorInit ((DbVisitor) object,
                 (DbVisitorFunction) NULL, 
                 (DbVisitorFunction) down, 
                 (DbVisitorFunction) in, 
                 (DbVisitorFunction) up);

  object->tab = 0;

  return ((DbVisitor) object);
}

/*---------------------------------------------------------------*/
static dump (DbObject object)
/*---------------------------------------------------------------*/
{
  char* type_name;
  char* name;
  DbVisitor visitor;

  type_name = DbTypeName (DbObjectGetType (object));
  name = DbObjectGetName (object);
  printf ("%s %s = ", type_name, name);

  visitor = DumperNew ();
  DbObjectVisit (object, visitor);
  DbVisitorDelete (visitor);

  printf ("\n");
}

/*---------------------------------------------------------------*/
int main (int argc, char* argv[])
/*---------------------------------------------------------------*/
{
  int do_scalars = 0;
  int do_vectors = 0;
  int do_matrices = 0;
  int do_structures = 0;
  DbReference* ref;

  int number = 0;

  if (!CmMessageOpenMultipleServer ("DbT"))
    {
      fprintf (stderr, "Bad Cm connection\n");
      exit (0);
    }

  if (argc > 1)
    {
      char* arg = argv[1];

      if (strchr (arg, 's'))
        {
          do_scalars = 1;
        }
      if (strchr (arg, 'v'))
        {
          do_vectors = 1;
        }
      if (strchr (arg, 'm'))
        {
          do_matrices = 1;
        }
      if (strchr (arg, 't'))
        {
          do_structures = 1;
        }

      if (argc > 2)
        {
          arg = argv[2];

          sscanf (arg, "%d", &number);
        }
    }
  else
    {
      do_scalars = 1;
      do_vectors = 1;
      do_matrices = 1;
      do_structures = 1;
    }

  ref = DbReferenceNew ();
    /* Scalars */

  if (do_scalars)
    {
      if ((number == 1) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/c");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 2) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/s");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 3) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/i");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 4) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/f");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 5) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/d");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 6) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/t");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 7) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/r");
          dump (DbReferenceGetObject (ref));
        }
    }

    /* Vectors */

  if (do_vectors)
    {
      if ((number == 1) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/vc");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 2) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/vs");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 3) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/vi");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 4) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/vf");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 5) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/vd");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 6) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/vt");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 7) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/vr");
          dump (DbReferenceGetObject (ref));
        }
    }
  
    /* Matrices */

  if (do_matrices)
    {
      if ((number == 1) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/mc");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 2) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/ms");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 3) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/mi");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 4) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/mf");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 5) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/md");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 6) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/mt");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 7) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/mr");
          dump (DbReferenceGetObject (ref));
        }
    }
  
    /* Structures */

  if (do_structures)
    {
      if ((number == 1) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/a");
          dump (DbReferenceGetObject (ref));
        }
      if ((number == 2) || (number == 0))
        {
          DbReferenceLoad (ref, "Virgo/va");
          dump (DbReferenceGetObject (ref));
        }
    }

  DbReferenceDelete (ref);

  DbPrintError ();

  DbCleanup ();
  CmMessageCleanup ();
    /*
  GCDumpBase ();
  */

  return (0);
}
